import json
import logging
import os

from ganzo.resolvers import resolve_templates
from ganzo.sources.gcs import GCSSource
from ganzo.sources.git import GitSourceWrapper
from ganzo.sources.local import LocalSource

logger = logging.getLogger(__name__)

GANZO_CONFIG_DIR = os.path.join("~", ".ganzo")
GANZO_CONFIG_FILE = "configuration.json"


def load_template_gcs_bucket_name():
    ganzo_home = os.path.expanduser(GANZO_CONFIG_DIR)
    config_file_path = os.path.join(ganzo_home, GANZO_CONFIG_FILE)

    if not os.path.isdir(ganzo_home):
        raise ValueError(f"No configuration folder available at '{ganzo_home}'.")

    if not os.path.isfile(config_file_path):
        raise ValueError(f"No configuration file available at '{config_file_path}'.")

    with open(config_file_path, "r", encoding="utf8") as config_file:
        config = json.load(config_file)

        if "gcs_bucket_name" in config:
            return config["gcs_bucket_name"]

        raise ValueError(
            f"Configuration attribute 'gcs_bucket_name' not defined in '{config_file_path}'."
        )


def list_templates(options):
    if options.local_source:
        source = LocalSource(options.local_source)
    else:
        gcs_bucket_name = load_template_gcs_bucket_name()
        source = GCSSource(gcs_bucket_name)

    if not options.legacy:
        source = GitSourceWrapper(source)

    templates = source.list_templates()
    for template in templates:
        print(template)


def load_template(options):
    if options.local_source:
        source = LocalSource(options.local_source)
    else:
        gcs_bucket_name = load_template_gcs_bucket_name()
        source = GCSSource(gcs_bucket_name)

    if not options.legacy:
        source = GitSourceWrapper(source)

    source.load_template(options.template, options.directory)
    resolve_templates(options.directory, {"__PROJECT_NAME__": options.project_name})



import shutil
import tempfile
from distutils.dir_util import copy_tree
from typing import List

from git.repo.base import Repo

from ganzo.sources import TemplateSource
from ganzo.sources.gcs import GCSSource
from ganzo.sources.local import LocalSource


class GitSourceWrapper(TemplateSource):
    def __init__(self, source: TemplateSource):
        self.source = source
        if isinstance(source, GCSSource) or isinstance(source, LocalSource):
            source.template_list_file_name = "templates-git.list"
        else:
            raise ValueError(f"{source} not supported")

    def list_templates(self) -> List[str]:
        return self.source.list_templates()

    def load_template(
        self,
        template_name: str,
        target_path: str,
    ):
        entries = [template.split(" ") for template in self.list_templates()]
        templates = {entry[0]: (entry[1], entry[2]) for entry in entries}

        if template_name not in templates:
            raise ValueError(f"{template_name} not an available template")

        with tempfile.TemporaryDirectory() as tmp_dir:
            print(f"Cloning {templates[template_name][0]} into '{tmp_dir}'")

            Repo.clone_from(
                templates[template_name][0],
                tmp_dir,
                multi_options=[
                    "--single-branch",
                    f"-b {templates[template_name][1]}",
                ],
            )

            shutil.rmtree(f"{tmp_dir}/.git")

            print(f"Copying {template_name} from '{tmp_dir}' into '{target_path}'")
            copy_tree(tmp_dir, target_path)



from abc import abstractmethod
from typing import List


class TemplateSource:
    @abstractmethod
    def list_templates(self) -> List[str]:
        raise NotImplementedError()

    @abstractmethod
    def load_template(self, template_name: str, target_path: str) -> None:
        raise NotImplementedError()



import os
import shutil
from pathlib import Path
from unittest.mock import Mock, patch

from ganzo.cli import run

_tests_root = Path(__file__).parent


def _cleanup(directory: str):
    shutil.rmtree(os.path.abspath(directory))


@patch("builtins.print")
def test_local_list_cli(mock_print: Mock):
    # Given
    templates_path = str(_tests_root.joinpath("fixtures", "templates"))
    args = ["--local-source", templates_path, "--legacy", "list"]

    # When
    run(args)

    # Then
    actuals = list(map(lambda c: c[0][0], mock_print.call_args_list))
    expecteds = ["example"]
    assert len(actuals) == len(expecteds)
    assert all(actual == expected for actual, expected in zip(actuals, expecteds))


def test_local_load_cli():
    # Given
    templates_path = str(_tests_root.joinpath("fixtures", "templates"))
    project_path = str(_tests_root.joinpath("test_project_1"))
    args = [
        "--local-source",
        templates_path,
        "--legacy",
        "load",
        "example",
        project_path,
        "test_project",
    ]

    # When
    run(args)

    # Then
    assert os.path.exists(project_path)
    assert os.path.exists(os.path.join(project_path, "some_file.txt"))
    assert os.path.exists(os.path.join(project_path, "another_file.txt"))

    _cleanup(project_path)


@patch("builtins.print")
def test_local_git_list_cli(mock_print: Mock):
    # Given
    templates_path = str(_tests_root.joinpath("fixtures", "templates"))
    args = ["--local-source", templates_path, "list"]

    # When
    run(args)

    # Then
    actuals = list(map(lambda c: c[0][0], mock_print.call_args_list))
    expecteds = ["example https://github.com/521xueweihan/HelloGitHub.git master"]
    assert len(actuals) == len(expecteds)
    assert all(actual == expected for actual, expected in zip(actuals, expecteds))


def test_local_git_load_cli():
    # Given
    templates_path = str(_tests_root.joinpath("fixtures", "templates"))
    project_path = str(_tests_root.joinpath("test_project_3"))
    args = [
        "--local-source",
        templates_path,
        "load",
        "example",
        project_path,
        "test_project",
    ]

    # When
    run(args)

    # Then
    assert os.path.exists(project_path)
    assert os.path.exists(os.path.join(project_path, "README.md"))
    assert os.path.exists(os.path.join(project_path, ".gitignore"))
    assert not os.path.exists(os.path.join(project_path, ".git"))

    _cleanup(project_path)


@patch("builtins.print")
def test_gcp_list_cli(mock_print: Mock):
    # Given
    args = ["--legacy", "list"]

    # When
    run(args)

    # Then
    actuals = list(map(lambda c: c[0][0], mock_print.call_args_list))
    expecteds = ["app", "lib"]
    assert all(actual == expected for actual, expected in zip(actuals, expecteds))


def test_gcp_load_cli():
    # Given
    project_path = str(_tests_root.joinpath("test_project_2"))
    args = ["--legacy", "load", "app", project_path, "test_project"]

    # When
    run(args)

    # Then
    assert os.path.exists(project_path)
    assert os.path.exists(os.path.join(project_path, "README.md"))

    _cleanup(project_path)



import os
from string import Template
from typing import Mapping


def resolve_templates(target_path: str, patterns: Mapping[str, str]):
    with os.scandir(target_path) as entries:
        for entry in entries:
            if entry.is_dir():
                resolve_templates(entry.path, patterns)
            if entry.is_file():
                _resolve_and_replace_file(entry.path, patterns)


def _resolve_and_replace_file(file_path: str, patterns: Mapping[str, str]):
    resolved_file_path, file_ext = os.path.splitext(file_path)
    if file_ext == ".nzo":
        print(f"Resolving '{file_path}' into '{resolved_file_path}'")
        with open(file_path, "r", encoding="utf-8") as source_file:
            file_content = source_file.read()
            content_template = Template(file_content)
            resolved_content = content_template.safe_substitute(**patterns)
            with open(resolved_file_path, "w", encoding="utf-8") as resolved_file:
                resolved_file.write(resolved_content)
        os.remove(file_path)



import os

from google.cloud import storage

from ganzo.sources import TemplateSource


# Because of Python <=3.8
def _remove_prefix(text: str, prefix: str) -> str:
    return text[len(prefix) :] if text.startswith(prefix) else text


class GCSSource(TemplateSource):
    def __init__(
        self,
        bucket_name: str,
        template_list_file_name: str = "templates.list",
    ):
        self.client = storage.Client()
        self.bucket_name = bucket_name
        self.template_list_file_name = template_list_file_name

    def list_templates(self):
        bucket = self.client.bucket(self.bucket_name)
        blob = bucket.blob(self.template_list_file_name)
        content = blob.download_as_bytes().decode("utf-8")
        return content.strip().split("\n")

    def load_template(self, template_name: str, target_path: str):
        template_path = f"{template_name}/"
        blobs = self.client.list_blobs(self.bucket_name, prefix=template_path)

        empty = True
        for blob in blobs:
            empty = False
            file_relative_path = _remove_prefix(blob.name, template_path)
            file_path = os.path.join(target_path, file_relative_path)
            file_dir, _ = os.path.split(file_path)

            print(f"Loading 'gs://{blob.bucket.name}/{blob.name}' into '{file_path}'")
            os.makedirs(file_dir, exist_ok=True)
            blob.download_to_filename(file_path)

        if empty:
            raise ValueError(f"Template '{template_name}' does not exist.")



import os
from distutils.dir_util import copy_tree
from typing import List

from ganzo.sources import TemplateSource


class LocalSource(TemplateSource):
    def __init__(
        self,
        base_path: str,
        template_list_file_name: str = "templates.list",
    ):
        self.base_path = base_path
        self.template_list_file_name = template_list_file_name

    def list_templates(self) -> List[str]:
        templates_path = os.path.join(self.base_path, self.template_list_file_name)
        with open(templates_path, "r", encoding="utf-8") as file:
            return file.read().strip().split("\n")

    def load_template(self, template_name: str, target_path: str):
        template_path = os.path.join(self.base_path, template_name)

        print(f"Copying '{template_path}' into '{target_path}'")
        copy_tree(template_path, target_path)



import argparse
import sys

from ganzo.core import list_templates, load_template


def run(args=None):
    parser = argparse.ArgumentParser(prog="ganzo")
    sub_parsers = parser.add_subparsers(help="Available commands")
    parser.add_argument(
        "-l",
        "--local-source",
        metavar="LOCAL_TEMPLATES_PATH",
        help="Load templates from local directory.",
    )

    parser.add_argument(
        "--legacy",
        default=False,
        action="store_true",
        help="Load templates using legacy strategy.",
    )

    list_parser = sub_parsers.add_parser("list", help="List available templates")
    list_parser.set_defaults(func=list_templates)

    create_parser = sub_parsers.add_parser(
        "load", help="Load template into target directory"
    )
    create_parser.set_defaults(func=load_template)
    create_parser.add_argument(
        "template",
        metavar="TEMPLATE_NAME",
        help="The name of the template to load.",
    )
    create_parser.add_argument(
        "directory", metavar="DIR_PATH", help="A directory to load the template into."
    )
    create_parser.add_argument(
        "project_name", metavar="PROJECT_NAME", help="Name of the project."
    )

    try:
        options = parser.parse_args(args)
        options.func(options)
    except (AttributeError, ValueError) as exception:
        print("Error:", exception, file=sys.stderr)
        parser.print_help()


def main():
    return_code = 1
    try:
        run()
        return_code = 0
    except Exception as exception:  # pylint: disable=W0703
        print(f"Error: {exception}", file=sys.stderr)
    sys.exit(return_code)


if __name__ == "__main__":
    main()
